//============================================================================
//  aPCI-8815 for PCI Linux Device Driver
//  apci8815.c
//  Author  Shoichiro Ito(ADTEK SYSTEM SCIENCE Co.,Ltd.)
//  Date    2001/03/16
//  Rev.    1.0.0
//============================================================================

#ifndef MODULE
#define MODULE
#endif // MODULE

#ifndef __KERNEL__
#define __KERNEL__
#endif // __KERNEL__

#include <linux/autoconf.h>
#if defined(CONFIG_MODVERSIONS) && !defined(MODVERSIONS)
#	define MODVERSIONS
#endif

#ifdef MODVERSIONS
#	include <linux/modversions.h>
#endif

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/fs.h>		// used file oprations
#include <linux/pci.h>		// used pci functions
#include <asm/io.h>		// used inb, outb
#include <asm/uaccess.h>	// used verify_area
#include <linux/poll.h>		// used poll_wait
#include <asm/delay.h>		// used udelay

#include "apci8815.h"

/*
 * Split minors in two parts
 */
#define TYPE(dev)		(MINOR(dev) >> 4)	// high nibble
#define NUM(dev)		(MINOR(dev) & 0xf)	// low	nibble


#define APCI8815_NAME		"apci8815"

#define APCI8815_VENDOR_ID	0x136C		// ǥХID,٥ID פʤ
#define APCI8815_DEVICE_ID	0x8815		// ɥ饤ФϿǤʤΤ

#define APCI8815_INT_OCCURS	1		// ȯ
#define APCI8815_INT_READY	0		// ȯԤ

#define APCI8815_DELAY_CH	45		// ͥؤλ

#ifdef APCI8815_DEBUG
#	define APCI8815_DELAY_CLOSE	1000 	// Ԥ
#else
#	define APCI8815_DELAY_CLOSE	200 	// Ԥ
#endif

#define PWARNING(fmt, args...)	printk(KERN_WARNING "[APCI8815]: " fmt, ## args)


/*----------------- Define Device Information ------------------------------*/

typedef struct _elmnt
{
	unsigned short data;
	char	       fvalid;
} elmnt, *pelmnt;

typedef struct _devinf
{
	unsigned int		io_addr;	// I/O Base Address
	unsigned char		irq;		// Interrupt Request
	unsigned long		f_version;	// ե뼱̻
	struct wait_queue*	wq;		// Ե
	int			int_occurs;	// ȯ

	unsigned short		ad_ctl;		// ץ󥰥ȥ쥸
	unsigned short		ad_md;		// ץ󥰥⡼
	unsigned long		ad_sts;		// ץ󥰥ơ
	unsigned short		ad_prev_ch;	// Υͥ

	pelmnt			pbuf;		// ץ󥰥ǡ
	pelmnt			pread_pos;	// ɤ߹߻ȥɥ쥹
	pelmnt			pwrite_pos;	// 񤭹߻ȥɥ쥹
	int			buf_cnt;	// Хåե
	int			data_cnt;	// Хåեǡ
	int			read_cnt;	// ɤ߹ߥ
	int			write_cnt;	// 񤭹ߥ
} devinf, *pdevinf;

static devinf	devctl[APCI8815_MAX_DEVICES];	// ǥХ
static unsigned int	uapci8815major;		// Major Device Number

static int devmajor = 0;


/*----------------- File Operations ----------------------------------------*/

static ssize_t      apci8815_read(struct file* file, char* buf, size_t count, loff_t* f_pos);
static unsigned int apci8815_poll(struct file* file, struct poll_table_struct* ptab);
static		int apci8815_ioctl(struct inode*, struct file*, unsigned int, unsigned long);
static		int apci8815_open(struct inode* inode, struct file* file);
static		int apci8815_close(struct inode* inode, struct file* file);

static struct file_operations apci8815_fops =
{
	NULL,		// loff_t	llseek(struct file*, loff_t, int)
	apci8815_read,	// ssize_t	read(struct file*, char*, size_t, loff_t*)
	NULL,		// ssize_t	write(struct file*, const char*, size_t, loff_t*)
	NULL,		// int		readdir(struct file*, void*, filldir_t)
	apci8815_poll,	// unsigned int poll(struct file*, struct poll_table_struct*)
	apci8815_ioctl,	// int		ioctl(struct inode*, struct file*, unsigned int, unsigned long)
	NULL,		// int		mmap(struct file*, struct vm_area_struct*)
	apci8815_open,	// int		open(struct inode*, struct file*)
	NULL,		// int		flush(struct file*)
	apci8815_close,	// int		release(struct inode*, struct file*)
	NULL,		// int		fsync(struct file*, struct dentry*)
	NULL,		// int		fasync(int, struct file*, int)
	NULL,		// int		check_media_change(kdev_t dev)
	NULL,		// int		revalidate(kdev_t dev)
	NULL,		// int		lock(struct file*, int, struct file_lock*)
};


/*----------------- Define Static Functions --------------------------------*/

static int		init_driver_info(void);
static int		setup_driver(void);
static void		release_driver(void);
static int		get_device(void);
static unsigned int	get_device_base_addr(unsigned int base_addr);
static int		regist_char_driver(void);
static int		unregist_char_driver(void);
static int		get_device_state(int ref);

static int	setup_hw_isr(int ref);
static void	release_hw_isr(int ref);

static int	verify_arg(unsigned int cmd, unsigned long arg);

static int	apci8815_ioc_set_ad_md(struct file* file, unsigned long arg);
static int	apci8815_ioc_start_ad(struct file* file, unsigned long arg);
static int	apci8815_ioc_stop_ad(struct file* file, unsigned long arg);
static int	apci8815_ioc_get_ad_sts(struct file* file, unsigned long arg);
static int	apci8815_ioc_get_exin(struct file* file, unsigned long arg);

static void	apci8815_isr(int irq, void* dev_id, struct pt_regs* regs);

static int      set_ad_md_01(int ref, papci8815_admd pmd);
static int	set_ad_md_02(int ref, papci8815_admd pmd);
static int	set_ad_md_03(int ref, papci8815_admd pmd);
static int	set_ad_md_04(int ref, papci8815_admd pmd);
static void	set_ad_md_common(int ref, papci8815_admd pmd);

static int	is_adbusy(int ref);
static int      is_channel(papci8815_admd pmd);
static int	is_cycle(papci8815_admd pmd);
static int	is_ovrun(int ref, unsigned short ch);
static unsigned short timer_to_twdata(int ref, papci8815_admd pmd);

static inline unsigned short ad_get_str(int ref);
static inline unsigned short ad_get_cd(int ref);
static inline void	     ad_set_ctl(int ref, unsigned short ctl);
static inline void	     ad_set_tw(int ref, unsigned short tw);
static inline void	     ad_set_intc(int ref);

static int  ring_alloc(int ref, int num);
static void ring_free(int ref);
static int  ring_read(int ref, unsigned short* pbuf, int cnt);
static int  ring_write(int ref, unsigned short data);
static void ring_init(int ref);


MODULE_PARM(devmajor, "i");	// ᥸㡼ֹѥ᡼


/*==========================================================================*/
/*  init_module - ɥ饤ФϿ					    */
/*==========================================================================*/
int init_module(void)
{
	int result;

	// PCI ǽ̵ͭȽ
	if (!pci_present())
		return -ENODEV;

	// ǥХɥ饤ѿν
	result = init_driver_info();
	if (result < 0)
		return result;

	// ǥХɥ饤ФΥåȥå
	result = setup_driver();
	if (result < 0)
		return result;

	return 0;
}

/*==========================================================================*/
/*  cleanup_module - ɥ饤ФϿä				    */
/*==========================================================================*/
void cleanup_module(void)
{
	// ǥХɥ饤Фβ
	release_driver();
}

/*--------------------------------------------------------------------------*/
/*  apci8815_read - ץ󥰥ǡɤ߹			    */
/*--------------------------------------------------------------------------*/
static ssize_t apci8815_read(struct file* file, char* buf, size_t count, loff_t* f_pos)
{
	int ref, res_cnt;

	PDEBUG("Read\n");

	// ǡλȰ
	ref = (int)file->private_data;

	// ץ󥰤λƤʤʤȴ
	if (devctl[ref].pread_pos->fvalid == 0)
		return 0;

	// 󥰥Хåեǡɤ߹
	res_cnt = ring_read(ref, (unsigned short*)buf, count);

	PDEBUG("data count=%d\n", devctl[ref].data_cnt);

	return res_cnt;
}

/*--------------------------------------------------------------------------*/
/*  apci8815_poll - ǥХߤΥݡ			    */
/*--------------------------------------------------------------------------*/
static unsigned int apci8815_poll(struct file* file, struct poll_table_struct* ptab)
{
	int ref;

	PDEBUG("Poll\n");

	// ǡλȰ
	ref = (int)file->private_data;

	// ȯ
	if (devctl[ref].int_occurs == APCI8815_INT_OCCURS)
	{
		devctl[ref].int_occurs = APCI8815_INT_READY;	// ȯԤذܹ
		return POLLIN | POLLRDNORM;
	}

	poll_wait(file, &devctl[ref].wq, ptab);

	return 0;
}

/*--------------------------------------------------------------------------*/
/*  apci8815_ioctl - ǥХͭ					    */
/*--------------------------------------------------------------------------*/
static int apci8815_ioctl(struct inode* inode, struct file* file, unsigned int cmd, unsigned long arg)
{
	int err = verify_arg(cmd, arg);
	if (err)
		return err;

	switch (cmd)
	{
		case APCI8815_SET_AD_MD:
			return apci8815_ioc_set_ad_md(file, arg);

		case APCI8815_START_AD:
			return apci8815_ioc_start_ad(file, arg);

		case APCI8815_STOP_AD:
			return apci8815_ioc_stop_ad(file, arg);

		case APCI8815_GET_AD_STS:
			return apci8815_ioc_get_ad_sts(file, arg);

		case APCI8815_GET_EXIN:
			return apci8815_ioc_get_exin(file, arg);
	}

	return 0;
}

/*--------------------------------------------------------------------------*/
/*  apci8815_open - ǥХΥץ					    */
/*--------------------------------------------------------------------------*/
static int apci8815_open(struct inode* inode, struct file* file)
{
	int ref;
	int result;

	PDEBUG("Open !! %ld\n", file->f_version);

	// Ȥ
	ref = NUM(inode->i_rdev);

	PDEBUG("ref index=%d\n", ref);

	// ǥХξ
	result = get_device_state(ref);
	if (result < 0)
		return result;

	// ߤ򤹤
	result = setup_hw_isr(ref);
	if (result < 0)
		return result;

	// ߴϢѿ
	devctl[ref].wq = NULL;
	devctl[ref].int_occurs = APCI8815_INT_READY;

	// Ѿ֤ˤ
	devctl[ref].f_version = file->f_version;

	// ǡȰ¸
	file->private_data = (void*)ref;

	MOD_INC_USE_COUNT;		// ȥ+

	return 0;
}

/*--------------------------------------------------------------------------*/
/*  apci8815_close - ǥХΥ					    */
/*--------------------------------------------------------------------------*/
static int apci8815_close(struct inode* inode, struct file* file)
{
	int ref;

	PDEBUG("Close !!\n");

	// ǡλȰ
	ref = (int)file->private_data;

	// ѿʤɤΥ꡼󥢥å׽

	// CTL ݡȥꥻå(ץʤ鶯λ)
	ad_set_intc(ref);

	devctl[ref].ad_ctl = APCI8815_CTL_RST;
	ad_set_ctl(ref, devctl[ref].ad_ctl);
	
	// ˳ߤäΤȤθ
	// ΥȤ
	udelay(APCI8815_DELAY_CLOSE);

	// ȯѿꥻå
	devctl[ref].int_occurs = APCI8815_INT_READY;

	// Ԥ󤬤ʤ
	while (waitqueue_active(&devctl[ref].wq))
		wake_up(&devctl[ref].wq);

	devctl[ref].ad_md  = 0;		// ץ󥰥⡼
	devctl[ref].ad_sts = 0;		// ץ󥰥ơ

	// ХåեݤƤʤ
	ring_free(ref);

	// ߤ
	release_hw_isr(ref);

	// ̤Ѿ֤ˤ
	devctl[ref].f_version = 0;

	MOD_DEC_USE_COUNT;		// ȥ-

	return 0;
}

/*--------------------------------------------------------------------------*/
/*  init_driver_info - ǥХɥ饤дѿν			    */
/*  Input : none							    */
/*  Output: return - 0 					    */
/*--------------------------------------------------------------------------*/
static int init_driver_info(void)
{
	int i;

	for (i = 0; i < APCI8815_MAX_DEVICES; i++)
	{
		devctl[i].io_addr    = 0x0000;
		devctl[i].irq	     = 0;
		devctl[i].f_version  = 0;

		devctl[i].ad_ctl     = APCI8815_CTL_RST;
		devctl[i].ad_md      = 0;
		devctl[i].ad_sts     = 0;

		devctl[i].pbuf       = NULL;
		ring_free(i);
	}

	return 0;
}

/*--------------------------------------------------------------------------*/
/*  setup_driver - ǥХɥ饤ФΥåȥå			    */
/*  Input : none							    */
/*  Output: return - 0 ʳ åȥå׼				    */
/*		     0 åȥå					    */
/*--------------------------------------------------------------------------*/
static int setup_driver(void)
{
	int result;

	result = get_device();
	if (result < 0)
		return result;

	// character device driver registration
	result = regist_char_driver();
	if (result < 0)
		return result;

	return 0;
}

/*--------------------------------------------------------------------------*/
/*  release_driver -  ǥХɥ饤Фβ				    */
/*  Input : none							    */
/*  Output: none							    */
/*--------------------------------------------------------------------------*/
static void release_driver(void)
{
	// character device driver unregistration
	unregist_char_driver();
}

/*--------------------------------------------------------------------------*/
/*  get_device - 󥹥ȡ뤵ƤǥХ		    */
/*  Input : none							    */
/*  Output: return - -ENODEV ǥХϸդʤ			    */
/*		     0	     ǥХϸդä			    */
/*--------------------------------------------------------------------------*/
static int get_device(void)
{
	struct pci_dev* pdev = NULL;

	unsigned int  io_addr;		// I/O Base Address
	unsigned char irq;		// Interrupt Request

	int i;

	for (i = 0; i < APCI8815_MAX_DEVICES; i++)
	{
		// PCI ǥХ
		pdev = pci_find_device(APCI8815_VENDOR_ID,
				       APCI8815_DEVICE_ID, pdev);
		if (pdev == NULL)
		    break;

		//  I/O ١ɥ쥹
		io_addr	  = get_device_base_addr(pdev->base_address[2]);
		// IRQ μ
		irq	  = pdev->irq;

		PDEBUG("io_addr:   %X\n", io_addr);
		PDEBUG("IRQ:      %d\n", irq);

		// ǡ
		devctl[i].io_addr = io_addr;	// I/O Base Address save
		devctl[i].irq	  = irq;	// Interrupt Request save
	}

	if (i == 0)
	{
		PWARNING("Not found Device(s)\n");
		return -ENODEV;
	}

	return 0;
}

/*--------------------------------------------------------------------------*/
/*  get_device_base_addr - ѲǽʥǥХΥ١ɥ쥹	    */
/*  Input : base_addr - ١ɥ쥹					    */
/*  Output: return - Խ١ɥ쥹				    */
/*--------------------------------------------------------------------------*/
static unsigned int get_device_base_addr(unsigned int base_addr)
{
	// I/O ١ɥ쥹쥸
	if ((base_addr & PCI_BASE_ADDRESS_SPACE) == PCI_BASE_ADDRESS_SPACE_IO)
		return base_addr & PCI_BASE_ADDRESS_IO_MASK;

	// Memory ١ɥ쥹쥸
	return base_addr & PCI_BASE_ADDRESS_MEM_MASK;
}

/*--------------------------------------------------------------------------*/
/*  regist_char_driver - ᥸㡼ֹưŪƤȥɥ饤ФϿ	    */
/*  Input : none							    */
/*  Output: return - register_chrdev 				    */
/*		     0	Ͽ					    */
/*--------------------------------------------------------------------------*/
static int regist_char_driver(void)
{
	int result;

	PDEBUG("devmajor=%d\n", devmajor);

	if (devmajor == 0)
	{
		// ᥸㡼ֹ椬ꤵƤʤ
		// ưŪ᥸㡼ֹμ
		result = register_chrdev(0, APCI8815_NAME, NULL);
		if (result < 0)
		{
			PDEBUG("get major device number faild.\n");
			return result;
		}

		uapci8815major = result;
	}
	else
	{
		// ᥸㡼ֹ椬ꤵƤ
		uapci8815major = devmajor;
	}

	PDEBUG("major number=%u\n", uapci8815major);

	// 饯ǥХɥ饤ФϿ
	result = register_chrdev(uapci8815major, APCI8815_NAME, &apci8815_fops);
	if (result < 0)
	{
		PDEBUG("register_chrdev faild.\n");
		return result;
	}

	return 0;
}

/*--------------------------------------------------------------------------*/
/*  unregist_char_driver - ɥ饤ФϿä			    */
/*  Input : none							    */
/*  Output: return - unregister_chrdev 				    */
/*		     0	Ͽ					    */
/*--------------------------------------------------------------------------*/
static int unregist_char_driver(void)
{
	int result;

	// 饯ǥХɥ饤ФϿ
	result = unregister_chrdev(uapci8815major, APCI8815_NAME);
	if (result < 0)
	{
		PDEBUG("unregister_chrdev faild.\n");
		return result;
	}

	return 0;
}

/*--------------------------------------------------------------------------*/
/*  get_device_state - ǥХξ֤				    */
/*  Input : ref	  - Ȱ						    */
/*  Output: return - -ENODEV ΰ֤˥ǥХϤʤ			    */
/*		     -EBUSY  					    */
/*		     0	     Ѳǽ 					    */
/*  Notes : ref >= 0 - ref <= 15 ޤǤʤ				    */
/*--------------------------------------------------------------------------*/
static int get_device_state(int ref)
{
	if (devctl[ref].io_addr == 0x0000)
		return -ENODEV;

	if (devctl[ref].f_version != 0)
		return -EBUSY;

	return 0;
}

/*--------------------------------------------------------------------------*/
/*  setup_hw_isr - ߥ롼				    */
/*  Input : ref	  - Ȱ						    */
/*  Output: return - ߥϥɥϿ				    */
/*--------------------------------------------------------------------------*/
static int setup_hw_isr(int ref)
{
	unsigned long flags = SA_INTERRUPT | SA_SHIRQ;	// ® | ͭ
	int result;

	// ߶ػ
	ad_set_ctl(ref, APCI8815_CTL_RST);

	// װ
	ad_set_intc(ref);

	// ߥϥɥ
	result = request_irq(devctl[ref].irq, apci8815_isr, flags, APCI8815_NAME, (void*)(unsigned)ref);
	if (result == 0)
		return 0;

	PDEBUG("request irq faild %d\n", devctl[ref].irq);

	return result;
}

/*--------------------------------------------------------------------------*/
/*  release_hw_isr - ߥ롼β				    */
/*  Input : ref - Ȱ						    */
/*  Output: none							    */
/*--------------------------------------------------------------------------*/
static void release_hw_isr(int ref)
{
	// ߥϥɥ
	free_irq(devctl[ref].irq, (void*)(unsigned)ref);
}

/*--------------------------------------------------------------------------*/
/*  verify_arg - ǥХͭΰ				    */
/*  Input : cmd - ioctl 楳ޥ					    */
/*	    arg - 桼֤ؤΥɥ쥹				    */
/*  Output: return - -EINVAL ̵ʰ					    */
/*		     verify_area 				    */
/*		     0	     Ѳǽ 					    */
/*--------------------------------------------------------------------------*/
static int verify_arg(unsigned int cmd, unsigned long arg)
{
	int size = _IOC_SIZE(cmd);

	/* don't even decode wrong cmds: better returning  EINVAL than EFAULT */
	if (_IOC_TYPE(cmd) != APCI8815_IOC_MAGIC)
		return -EINVAL;

	if (_IOC_NR(cmd) > APCI8815_IOC_MAXNR)
		return -EINVAL;

	/*
	 * the type is a bitmask, and VERIFY_WRITE catches R/W
	 * transfers. Note that the type is user-oriented, while
	 * verify_area is kernel-oriented, so the concept of "read" and
	 * "write" is reversed
	 */
	if (_IOC_TYPE(cmd) & _IOC_READ)
		return verify_area(VERIFY_WRITE, (void*)arg, size);

	if (_IOC_TYPE(cmd) & _IOC_WRITE)
		return verify_area(VERIFY_READ, (void*)arg, size);

	return 0;
}

/*--------------------------------------------------------------------------*/
/*  apci8815_ioc_set_ad_md - A/D ץ󥰥⡼ɤ		    */
/*  Input : file - ͥ file ¤ΤΥݥ			    */
/*	    arg	 - 桼֤ؤΥɥ쥹				    */
/*  Output: return - -EINVAL ̵ʰ					    */
/*		     -EBUSY  ǥХϥӥ			    */
/*		     -ENOMEM ­					    */
/*		     0	     ｪλ					    */
/*--------------------------------------------------------------------------*/
static int apci8815_ioc_set_ad_md(struct file* file, unsigned long arg)
{
	int ref, rslt;
	apci8815_admd md;

	PDEBUG("Set mode\n");

	// ǡλȰ
	ref = (int)file->private_data;

	// ץ󥰼¹
	if (is_adbusy(ref))
		return -EBUSY;

	// 桼֤ǡ
	copy_from_user(&md, (void*)arg, sizeof(md));

	// Хåե
	if (md.buf_size < 1)
		return -EBUSY;

	ring_free(ref);

	rslt = ring_alloc(ref, md.buf_size);
	if (rslt < 0)
		return rslt;

	switch (md.md)
	{
		case APCI8815_AD_MD_01:
			return set_ad_md_01(ref, &md);

		case APCI8815_AD_MD_02:
			return set_ad_md_02(ref, &md);

		case APCI8815_AD_MD_03:
			return set_ad_md_03(ref, &md);

		case APCI8815_AD_MD_04:
			return set_ad_md_04(ref, &md);

		default:
			return -EINVAL;
	}

	return 0;
}

/*--------------------------------------------------------------------------*/
/*  apci8815_ioc_start_ad - A/D ץ󥰤򳫻			    */
/*  Input : file - ͥ file ¤ΤΥݥ			    */
/*	    arg	 - 桼֤ؤΥɥ쥹				    */
/*  Output: return - -EINVAL ̵ʰ					    */
/*		     -EBUSY  ǥХϥӥ			    */
/*		     0	     ｪλ					    */
/*--------------------------------------------------------------------------*/
static int apci8815_ioc_start_ad(struct file* file, unsigned long arg)
{
	int ref;

	PDEBUG("Start Sampling\n");

	// ǡλȰ
	ref = (int)file->private_data;

	// ץ󥰼¹
	if (is_adbusy(ref))
		return -EBUSY;

	if (!(devctl[ref].ad_md == APCI8815_AD_MD_01 || devctl[ref].ad_md == APCI8815_AD_MD_02 ||
	      devctl[ref].ad_md == APCI8815_AD_MD_03 || devctl[ref].ad_md == APCI8815_AD_MD_04))
		return -EINVAL;

	devctl[ref].ad_prev_ch = APCI8815_PREV_CH_RESET;	// Υͥ륤󥯥ͥꥻå

	devctl[ref].ad_sts = 0;	// ơꥢ

	ring_init(ref);		// Хåեꥢ

	// A/D Ѵ
	devctl[ref].ad_ctl |= APCI8815_CTL_START;

	// CTL 
	ad_set_ctl(ref, devctl[ref].ad_ctl);

	devctl[ref].ad_sts = APCI8815_STS_SMP;

	return 0;
}

/*--------------------------------------------------------------------------*/
/*  apci8815_ioc_stop_ad - A/D ץ󥰤			    */
/*  Input : file - ͥ file ¤ΤΥݥ			    */
/*	    arg	 - 桼֤ؤΥɥ쥹				    */
/*  Output: return - 0	     ｪλ					    */
/*--------------------------------------------------------------------------*/
static int apci8815_ioc_stop_ad(struct file* file, unsigned long arg)
{
	int ref;

	PDEBUG("Stop Sampling\n");

	// ǡλȰ
	ref = (int)file->private_data;

	// A/D Ѵ
	devctl[ref].ad_ctl &= ~APCI8815_CTL_START;
	ad_set_ctl(ref, devctl[ref].ad_ctl);

	// ˳ߤäΤȤθ
	// ΥȤ
	udelay(APCI8815_DELAY_CLOSE);

	switch (devctl[ref].ad_md)
	{
		case APCI8815_AD_MD_01:
		case APCI8815_AD_MD_02:
			devctl[ref].ad_sts &= APCI8815_STS_SMP;
			break;

		case APCI8815_AD_MD_03:
		case APCI8815_AD_MD_04:
			devctl[ref].ad_sts &= APCI8815_STS_SMP;
			devctl[ref].ad_sts &= APCI8815_STS_SMP_OVRUN;
			break;
	}

	return 0;
}

/*--------------------------------------------------------------------------*/
/*  apci8815_ioc_get_ad_sts - A/D ץ󥰥ơ		    */
/*  Input : file - ͥ file ¤ΤΥݥ			    */
/*	    arg	 - 桼֤ؤΥɥ쥹				    */
/*  Output: return - 0	     ｪλ					    */
/*--------------------------------------------------------------------------*/
static int apci8815_ioc_get_ad_sts(struct file* file, unsigned long arg)
{
	int ref;

	PDEBUG("Get Status\n");

	// ǡλȰ
	ref = (int)file->private_data;

	// 桼֤إǡ
	put_user(devctl[ref].ad_sts, (unsigned long*)arg);

	return 0;
}

/*--------------------------------------------------------------------------*/
/*  apci8815_ioc_get_exin - ϥݡȤǡ		    */
/*  Input : file - ͥ file ¤ΤΥݥ			    */
/*	    arg	 - 桼֤ؤΥɥ쥹				    */
/*  Output: return - 0	     ｪλ					    */
/*--------------------------------------------------------------------------*/
static int apci8815_ioc_get_exin(struct file* file, unsigned long arg)
{
	int ref;
	unsigned short data;

	PDEBUG("Get EX IN\n");

	// ǡλȰ
	ref = (int)file->private_data;

	// STR ݡ
	data = ad_get_str(ref);

	// ɬפʾΤߤˤ
	data &= (APCI8815_STR_EXTRG | APCI8815_STR_EXIN0 | APCI8815_STR_EXIN1);

	// 桼֤إǡ
	put_user(data, (unsigned short*)arg);

	return 0;
}

/*--------------------------------------------------------------------------*/
/*  apci8815_isr - ߥӥ롼				    */
/*--------------------------------------------------------------------------*/
static void apci8815_isr(int irq, void* dev_id, struct pt_regs* regs)
{
	unsigned ref = (unsigned)dev_id;
	unsigned short str;
	unsigned short data, ch;

//	PDEBUG("Interrupt ref=%d\n", ref);

	// ߸
	if (!(ref >= 0 && ref < APCI8815_MAX_DEVICES))	// ϰ
		return;

	if (devctl[ref].f_version == 0)		// ǥХƤʤ
		return;

	str = ad_get_str(ref);

	// STR ݡȤɤǡCVDF bit ƥ֤ɤå
	// ƥ֤Ǥʤʤ顢ΥǥХФƤγߤǤϤʤ
	if ((str & APCI8815_STR_CVDF) == 0)
		return;

//	PDEBUG("Interrupt str=%02X\n", str);

	// CD ݡȤɤ߹(A/D Ѵǡɤ߹)
	data = ad_get_cd(ref);

	// INTC ݡȤإǡϡ׵βԤ
	ad_set_intc(ref);

	// ץ󥰥ǡȥͥ륢ɥ쥹
	ch = (data << 12) & 0xF000;
	data = (data >> 4) & 0x0FFF;
	data |= ch;

	// 󥰥Хåեդʤ
	if (devctl[ref].pwrite_pos == devctl[ref].pread_pos && devctl[ref].pread_pos->fvalid == 1)
	{
		devctl[ref].ad_sts &= ~APCI8815_STS_BUF_HFULL;
		devctl[ref].ad_sts &= ~APCI8815_STS_BUF_FULL;
		devctl[ref].ad_sts |= APCI8815_STS_BUF_OVFLOW;

		devctl[ref].int_occurs = APCI8815_INT_OCCURS;
		wake_up_interruptible(&devctl[ref].wq);

//		PDEBUG("Wake up Interrupt ovflow\n");

		// ȥꥬˤ륵ץ󥰤ʤƥȤ򤫤롣
		if (devctl[ref].ad_md == APCI8815_AD_MD_04)
			ad_set_ctl(ref, devctl[ref].ad_ctl);

		return;
	}

	// 󥰥륵ץ
	if (devctl[ref].ad_md == APCI8815_AD_MD_01 || devctl[ref].ad_md == APCI8815_AD_MD_02)
	{
		ring_write(ref, data);

		devctl[ref].ad_sts &= ~APCI8815_STS_SMP;
		devctl[ref].ad_sts |= APCI8815_STS_EOC;

		// A/D Ѵ
		devctl[ref].ad_ctl &= ~APCI8815_CTL_START;
		ad_set_ctl(ref, devctl[ref].ad_ctl);

		devctl[ref].int_occurs = APCI8815_INT_OCCURS;
		wake_up_interruptible(&devctl[ref].wq);

		PDEBUG("Wake up Interrupt eoc\n");

		return;
	}

	// ץ󥰥Хå
	if (is_ovrun(ref, ch))
	{
		// A/DѴȥå
		devctl[ref].ad_ctl &= ~APCI8815_CTL_START;
		ad_set_ctl(ref, devctl[ref].ad_ctl);

		// Сž OFF
		if (!(devctl[ref].ad_sts & APCI8815_STS_SMP_OVRUN))
		{
			devctl[ref].int_occurs = APCI8815_INT_OCCURS;
			wake_up_interruptible(&devctl[ref].wq);

			PDEBUG("Wake up Interrupt  ovrun\n");
		}

		devctl[ref].ad_sts = APCI8815_STS_SMP_OVRUN;
	}

	// ǡ¸
	ring_write(ref, data);

	// Хåեξ
	if (devctl[ref].buf_cnt == devctl[ref].data_cnt)
	{
		devctl[ref].int_occurs = APCI8815_INT_OCCURS;
		wake_up_interruptible(&devctl[ref].wq);

		PDEBUG("Wake up Interrupt buf full\n");

		devctl[ref].ad_sts &= ~APCI8815_STS_BUF_HFULL;
		devctl[ref].ad_sts |= APCI8815_STS_BUF_FULL;

	}
	else if (devctl[ref].buf_cnt / 2 <= devctl[ref].data_cnt)
	{
		if (!(devctl[ref].ad_sts & APCI8815_STS_BUF_HFULL))
		{
			devctl[ref].int_occurs = APCI8815_INT_OCCURS;
			wake_up_interruptible(&devctl[ref].wq);

			PDEBUG("Wake up Interrupt buf h full\n");
		}

		devctl[ref].ad_sts |= APCI8815_STS_BUF_HFULL;

	}
	else
	{
		devctl[ref].ad_sts &= ~APCI8815_STS_BUF_HFULL;
		devctl[ref].ad_sts &= ~APCI8815_STS_BUF_FULL;
		devctl[ref].ad_sts &= ~APCI8815_STS_BUF_OVFLOW;
	}

	// ȥꥬˤ륵ץ󥰤ʤƥȤ򤫤롣
	if (devctl[ref].ad_md == APCI8815_AD_MD_04)
		ad_set_ctl(ref, devctl[ref].ad_ctl);
}

/*--------------------------------------------------------------------------*/
/*  set_ad_md_01 - A/D ץ󥰥⡼ (01)			    */
/*  Input : ref	 - Ȱ						    */
/*	    pmd	 - ץ					    */
/*  Output: return - -EINVAL ̵ʰ					    */
/*		     0	     ｪλ					    */
/*  Note  : 󥰥륵ץ󥰡եȥȥꥬ				    */
/*--------------------------------------------------------------------------*/
static int set_ad_md_01(int ref, papci8815_admd pmd)
{
	if (!is_channel(pmd))
		return -EINVAL;

	// ⡼ɶ̽
	set_ad_md_common(ref, pmd);

	// CTL 
	ad_set_ctl(ref, devctl[ref].ad_ctl);

	udelay(APCI8815_DELAY_CH);

	devctl[ref].ad_md = pmd->md;

	return 0;
}

/*--------------------------------------------------------------------------*/
/*  set_ad_md_02 - A/D ץ󥰥⡼ (02)			    */
/*  Input : ref	 - Ȱ						    */
/*	    pmd	 - ץ					    */
/*  Output: return - -EINVAL ̵ʰ					    */
/*		     0	     ｪλ					    */
/*  Note  : 󥰥륵ץ󥰡ȥꥬ				    */
/*--------------------------------------------------------------------------*/
static int set_ad_md_02(int ref, papci8815_admd pmd)
{
	if (!is_channel(pmd))
		return -EINVAL;

	// ⡼ɶ̽
	set_ad_md_common(ref, pmd);

	// ȥꥬ
	devctl[ref].ad_ctl |= APCI8815_CTL_EXTSW;

	// CTL 
	ad_set_ctl(ref, devctl[ref].ad_ctl);

	udelay(APCI8815_DELAY_CH);

	devctl[ref].ad_md = pmd->md;

	return 0;
}

/*--------------------------------------------------------------------------*/
/*  set_ad_md_03 - A/D ץ󥰥⡼ (03)			    */
/*  Input : ref	 - Ȱ						    */
/*	    pmd	 - ץ					    */
/*  Output: return - -EINVAL ̵ʰ					    */
/*		     0	     ｪλ					    */
/*  Note  : Ϣ³ץ󥰡եȥȥꥬ				    */
/*--------------------------------------------------------------------------*/
static int set_ad_md_03(int ref, papci8815_admd pmd)
{
	unsigned short tw;

	if (!is_channel(pmd))
		return -EINVAL;

	if (!is_cycle(pmd))
		return -EINVAL;

	// ⡼ɶ̽
	set_ad_md_common(ref, pmd);

	// ޤ A/D Ѵǡΰץ
	devctl[ref].ad_ctl |= APCI8815_CTL_SMP;

	// ޤΥǡѴ
	tw = timer_to_twdata(ref, pmd);

	// CTL 
	ad_set_ctl(ref, devctl[ref].ad_ctl);

	udelay(APCI8815_DELAY_CH);

	// TW 
	ad_set_tw(ref, tw);

	devctl[ref].ad_md = pmd->md;

	return 0;
}

/*--------------------------------------------------------------------------*/
/*  set_ad_md_04 - A/D ץ󥰥⡼ (04)			    */
/*  Input : ref	 - Ȱ						    */
/*	    pmd	 - ץ					    */
/*  Output: return - -EINVAL ̵ʰ					    */
/*		     0	     ｪλ					    */
/*  Note  : Ϣ³ץ󥰡ȥꥬ				    */
/*--------------------------------------------------------------------------*/
static int set_ad_md_04(int ref, papci8815_admd pmd)
{
	if (!is_channel(pmd))
		return -EINVAL;

	// ⡼ɶ̽
	set_ad_md_common(ref, pmd);

	// ȥꥬ
	devctl[ref].ad_ctl |= APCI8815_CTL_EXTSW;

	// CTL 
	ad_set_ctl(ref, devctl[ref].ad_ctl);

	udelay(APCI8815_DELAY_CH);

	devctl[ref].ad_md = pmd->md;

	return 0;
}

/*--------------------------------------------------------------------------*/
/*  set_ad_md_common - A/D ץ󥰥⡼()			    */
/*  Input : ref	 - Ȱ						    */
/*	    pmd	 - ץ					    */
/*  Output: return - -EINVAL ̵ʰ					    */
/*		     0	     ｪλ					    */
/*--------------------------------------------------------------------------*/
static void set_ad_md_common(int ref, papci8815_admd pmd)
{
	unsigned short ch;

	// CTL ݡȥꥻå
	devctl[ref].ad_ctl = APCI8815_CTL_RST;
	ad_set_ctl(ref, devctl[ref].ad_ctl);

	// ͥ롦ȡ󥯥ȵ
	if (pmd->ch_first != pmd->ch_last)
		devctl[ref].ad_ctl |= APCI8815_CTL_CISW;

	// ͥ륢ɥ쥹
	ch = (unsigned short)((pmd->ch_last << 4) | pmd->ch_first);

	// CTL  15-8bit ž
	devctl[ref].ad_ctl |= (ch << 8);
}


/*--------------------------------------------------------------------------*/
/*  is_adbusy - A/D ¹Ծ						    */
/*  Input : ref	 - Ȱ						    */
/*  Output: return - 0 ¹						    */
/*		     1 ¹						    */
/*--------------------------------------------------------------------------*/
static int is_adbusy(int ref)
{
	return devctl[ref].ad_ctl & APCI8815_CTL_START;
}

/*--------------------------------------------------------------------------*/
/*  is_channel - ץ󥰳Ͻλͥå			    */
/*  Input : pmd	 - ץ					    */
/*  Output: return - 0 ϰϳ						    */
/*		     1 ϰ						    */
/*--------------------------------------------------------------------------*/
static int is_channel(papci8815_admd pmd)
{
	// ץ󥰳ϥͥ(0  15)
	if (pmd->ch_first < APCI8815_CH_FIRST || pmd->ch_first > APCI8815_CH_LAST)
		return 0;

	// ץ󥰽λͥ(0  15)
	if (pmd->ch_last < APCI8815_CH_FIRST || pmd->ch_last > APCI8815_CH_LAST)
		return 0;

	// {ch_first <= ch_last}
	if (pmd->ch_first > pmd->ch_last)
		return 0;

	return 1;
}

/*--------------------------------------------------------------------------*/
/*  is_cycle - ץ󥰥ޥå				    */
/*  Input : pmd	 - ץ					    */
/*  Output: return - 0 ϰϳ						    */
/*		     1 ϰ						    */
/*--------------------------------------------------------------------------*/
static int is_cycle(papci8815_admd pmd)
{
	// ͥ
	if (pmd->ch_first == pmd->ch_last)
		if (pmd->cycle < APCI8815_CYCL_MIN || pmd->cycle > APCI8815_CYCL_MAX)
			return 0;

	// ͥؤ
	if (pmd->ch_first != pmd->ch_last)
		if (pmd->cycle < APCI8815_CYCL_CI_MIN || pmd->cycle > APCI8815_CYCL_MAX)
			return 0;

	return 1;
}

/*--------------------------------------------------------------------------*/
/*  is_ovrun - Хå					    */
/*  Input : ref	 - Ȱ						    */
/*	    ch   - ߤΥץ󥰥ͥ				    */
/*  Output: return - 0 ʤ						    */
/*		     1 Хȯ					    */
/*--------------------------------------------------------------------------*/
static int is_ovrun(int ref, unsigned short ch)
{
	unsigned short data, ch_first, ch_last, ch_cur;

	// ͥ륤󥯥ȤǤϤʤ
	if (!(devctl[ref].ad_ctl & APCI8815_CTL_CISW))
		return 0;

	// CTL  7-0bit ž
	data = (devctl[ref].ad_ctl >> 8) & 0x00FF;
	ch_first = data & 0x000F;
	ch_last  = (data >> 4) & 0x000F;
	ch_cur = (ch >> 12) & 0x000F;

	// ϰ˼ޤäƤ
	if (ch_last < ch_cur)	// 곰Υץ󥰤Ԥ줿
		return 1;

	if (devctl[ref].ad_prev_ch == APCI8815_PREV_CH_RESET)	// Υͥ뤬ꥻå
	{
		devctl[ref].ad_prev_ch = ch_cur;
		return 0;
	}
	if (devctl[ref].ad_prev_ch == ch_last) // Υͥ뤬λͥʤ
	{
		if (ch_cur == APCI8815_NOW_CH_ZERO)	// ߤΥͥ뤬 0 ʤ OK
		{
			devctl[ref].ad_prev_ch = ch_cur;
			return 0;
		}
	}
	else							// Υͥ뤬λͥʳʤ
	{
		if (devctl[ref].ad_prev_ch == ch_cur - 1)	// -1ǰƱͤʤ OK
		{
			devctl[ref].ad_prev_ch = ch_cur;
			return 0;
		}
	}

	// 곰Υץ󥰤Ԥ줿
	return 1;
}

/*--------------------------------------------------------------------------*/
/*  is_ovrun - ޤΥǡѴ					    */
/*  Input : ref	 - Ȱ						    */
/*          pmd	 - ץ					    */
/*  Output: return - ޥǡѴ					    */
/*--------------------------------------------------------------------------*/
static unsigned short timer_to_twdata(int ref, papci8815_admd pmd)
{
	// 12		 999ξ硢͡ˡݣ᥿
	if (pmd->cycle >=	   12 && pmd->cycle <=		 999)
		return (unsigned short)(pmd->cycle / 4 - 1);

	//	 1000	  249999ξ硢͡1000ˡݣ᥿
	if (pmd->cycle >=	 1000 && pmd->cycle <=	  249999)
	{
		devctl[ref].ad_ctl |= APCI8815_CTL_E0;		// ϼȿ0
		return (unsigned short)(pmd->cycle / 1000 - 1);
	}

	// 250000 64000000ξ硢͡250000ˡݣ᥿
	if (pmd->cycle >= 250000 && pmd->cycle <= 64000000)
	{
		devctl[ref].ad_ctl |= APCI8815_CTL_E1;		// ϼȿ1
		return (unsigned short)(pmd->cycle / 250000 - 1);
	}

	return 0;	// ǰΤ
}

/*--------------------------------------------------------------------------*/
/*  ad_get_str - A/D ơ					    */
/*  Input : ref - Ȱ						    */
/*  Output: return - A/D ơ					    */
/*--------------------------------------------------------------------------*/
static inline unsigned short ad_get_str(int ref)
{
	return inw(devctl[ref].io_addr + APCI8815_STR_PORT);
}

/*--------------------------------------------------------------------------*/
/*  ad_get_str - A/D Ѵǡ					    */
/*  Input : ref - Ȱ						    */
/*  Output: return - A/D Ѵǡ					    */
/*--------------------------------------------------------------------------*/
static inline unsigned short ad_get_cd(int ref)
{
	return inw(devctl[ref].io_addr + APCI8815_CD_PORT);
}

/*--------------------------------------------------------------------------*/
/*  apic_set_icr - A/D ȥ					    */
/*  Input : ref	 - Ȱ						    */
/*	    ctl	 - 						    */
/*  Output: none							    */
/*--------------------------------------------------------------------------*/
static inline void ad_set_ctl(int ref, unsigned short ctl)
{
	outw(ctl, devctl[ref].io_addr + APCI8815_CTL_PORT);
}

/*--------------------------------------------------------------------------*/
/*  ad_set_tw - TW 							    */
/*  Input : ref	 - Ȱ						    */
/*	    tw	 - 						    */
/*  Output: none							    */
/*--------------------------------------------------------------------------*/
static inline void ad_set_tw(int ref, unsigned short tw)
{
	outw(tw, devctl[ref].io_addr + APCI8815_TW_PORT);
}

/*--------------------------------------------------------------------------*/
/*  ad_set_intc - ߥꥢ					    */
/*  Input : ref	 - Ȱ						    */
/*  Output: none							    */
/*--------------------------------------------------------------------------*/
static inline void ad_set_intc(int ref)
{
	// CTL ݡȥꥻå
	outw(APCI8815_INTC_RST, devctl[ref].io_addr + APCI8815_INTC_PORT);
}

/*--------------------------------------------------------------------------*/
/*  ring_alloc - ץѥ󥰥Хåեγ			    */
/*  Input : ref - Ȱ						    */
/*	    num - ݿ						    */
/*  Output: return - -ENOMEM ­					    */
/*		     0	     ｪλ					    */
/*--------------------------------------------------------------------------*/
static int ring_alloc(int ref, int num)
{
	devctl[ref].pbuf = vmalloc((num + 1) * sizeof(elmnt));
	if (devctl[ref].pbuf == NULL)
		return -ENOMEM;

	devctl[ref].buf_cnt = num;

	ring_init(ref);

	return 0;
}

/*--------------------------------------------------------------------------*/
/*  ring_free - ץѥ󥰥Хåեβ			    */
/*  Input : ref - Ȱ						    */
/*  Output: none							    */
/*--------------------------------------------------------------------------*/
static void ring_free(int ref)
{
	if (devctl[ref].pbuf != NULL)
		vfree(devctl[ref].pbuf);

	devctl[ref].pbuf       = NULL;
	devctl[ref].pread_pos  = NULL;
	devctl[ref].pwrite_pos = NULL;

	devctl[ref].buf_cnt    = 0;
	devctl[ref].data_cnt   = 0;
	devctl[ref].read_cnt   = 0;
	devctl[ref].write_cnt  = 0;
}

/*--------------------------------------------------------------------------*/
/*  ring_init - ץѥ󥰥Хåեν			    */
/*  Input : ref - Ȱ						    */
/*  Output: none							    */
/*--------------------------------------------------------------------------*/
static void ring_init(int ref)
{
	int i;

	for (i = 0; i < devctl[ref].buf_cnt; i++)
		devctl[ref].pbuf[i].fvalid = 0;

	devctl[ref].pread_pos = devctl[ref].pwrite_pos = devctl[ref].pbuf;
}

/*--------------------------------------------------------------------------*/
/*  ring_read - ץѥ󥰥Хåեǡ		    */
/*  Input : ref  - Ȱ						    */
/*	    pbuf - ǼХåեؤΥɥ쥹				    */
/*	    cnt  - ɤߤ						    */
/*  Output: return - ºݤɤߤ					    */
/*--------------------------------------------------------------------------*/
static int ring_read(int ref, unsigned short* pbuf, int cnt)
{
	int i;
	int num = 0;
	unsigned short* p = pbuf;

	if (devctl[ref].pbuf == NULL)
		return 0;

	for (i = 0; i < cnt; i++)
	{
		// ɤ߽ФǡϤʤ
		if (devctl[ref].pread_pos->fvalid != 1)
			break;

		put_user(devctl[ref].pread_pos->data, p);	// ǡž
		p++;						// ž
		devctl[ref].pread_pos->fvalid = 0;		// ǡ̵
		devctl[ref].pread_pos++;			// ž
		devctl[ref].data_cnt--;

		num++;
		if (devctl[ref].pread_pos >= &devctl[ref].pbuf[devctl[ref].buf_cnt])
			devctl[ref].pread_pos = devctl[ref].pbuf;
	}

	// ɤ߽Ф
	return num;
}

/*--------------------------------------------------------------------------*/
/*  ring_write - ץѥ󥰥Хåեإǡ		    */
/*  Input : ref  - Ȱ						    */
/*	    data - ǡ						    */
/*  Output: return - -1 Хåեϰ					    */
/*			 0  ｪλ					    */
/*--------------------------------------------------------------------------*/
static int ring_write(int ref, unsigned short data)
{
	// ХåեϰդˤʤäƤ
	if (devctl[ref].pwrite_pos == devctl[ref].pread_pos && devctl[ref].pread_pos->fvalid == 1)
		return -1;

	devctl[ref].pwrite_pos->data = data;
	devctl[ref].pwrite_pos->fvalid = 1;
	devctl[ref].pwrite_pos++;
	devctl[ref].data_cnt++;

	if (devctl[ref].pwrite_pos >= &devctl[ref].pbuf[devctl[ref].buf_cnt])
		devctl[ref].pwrite_pos = devctl[ref].pbuf;

	return 0;
}
