//============================================================================
//  aPCI-8815 Device Driver I/F Control
//  apci8815ctl.c
//  Author  Shoichiro Ito(ADTEK SYSTEM SCIENCE Co.,Ltd.)
//  Date    2001/03/16
//  Rev.    1.0.0
//============================================================================

#include <stdio.h>	// used printf,perror...
#include <pthread.h>	// used pthread
#include <fcntl.h>	// used open
#include <unistd.h>	// used close,usleep
#include <sys/ioctl.h>	// used ioctl
#include <sys/poll.h>	// used poll
#include <string.h>	// used strerror
#include <errno.h>	// used errno

#include "apci8815ctl.h"


/*----------------- Define Device Control ----------------------------------*/

#define APCI8815_FNC_EXEC	1		// ¹
#define APCI8815_FNC_NOTEXEC	0		// ¹Բǽ

typedef struct _apci8815_ctl
{
	int	  fd;
	pthread_t hisr;
	void	  (*pcallback)(int, void*);
	void*	  arg;
	int	  exit_isr;
	int	  bexec;			// ǽ¹
} apci8815_ctl, *papci8815_ctl;

static apci8815_ctl ctl[APCI8815_MAX_DEVICES] =
{
	{ 0 }, { 0 }, { 0 }, { 0 }, { 0 }, { 0 }, { 0 }, { 0 },
	{ 0 }, { 0 }, { 0 }, { 0 }, { 0 }, { 0 }, { 0 }, { 0 }
};


/*----------------- Define Static Functions --------------------------------*/

static int	dev_open(int log_slot);
static int	is_log_slot(int log_slot);
static int	is_pass(int log_slot);
static void*	thread_isr(void* arg);


/*==========================================================================*/
/*  apci8815_create - ǥХץ					    */
/*  Input : log_slot - åȻ(APCI8815_SLOT_AUTO ξ硢	    */
/*					̤ѥǥХ)		    */
/*  Output: return - 0 ̤ 顼				    */
/*		     0 ʾ å				    */
/*==========================================================================*/
int apci8815_create(int log_slot)
{
	int fd = 0, i;

	// åȻ
	if (log_slot != APCI8815_SLOT_AUTO)
	{
		// åȤ
		if (!(log_slot >= 0 && log_slot < APCI8815_MAX_DEVICES))
			return APCI8815_EINVAL;

		// ǥХץ
		fd = dev_open(log_slot);
		if (fd < 0)
		{
			PDEBUG("Error: open, %s\n", strerror(errno));

			if (errno == EBUSY)
				return APCI8815_EBUSY;
			return APCI8815_ENODEV;
		}

		ctl[log_slot].fd    = fd;
		ctl[log_slot].bexec = APCI8815_FNC_NOTEXEC;
		ctl[log_slot].hisr  = 0;

		return log_slot;
	}

	// ̤ѥǥХץ
	for (i = 0; i < APCI8815_MAX_DEVICES; i++)
	{
		// ǥХϥץ
		if (ctl[i].fd != 0)
			continue;

		// ǥХץ
		fd = dev_open(i);
		if (fd < 0)
			continue;

		break;
	}

	// ץǤǥХ̵
	if (i >= APCI8815_MAX_DEVICES)
		return APCI8815_ENODEV;

	ctl[i].fd    = fd;
	ctl[i].bexec = APCI8815_FNC_NOTEXEC;
	ctl[i].hisr  = 0;

	return i;		// return log_slot
}

/*==========================================================================*/
/*  apci8815_close - ǥХ					    */
/*  Input : log_slot - åȻ					    */
/*  Output: return - 0 ̤ 顼				    */
/*		     0 						    */
/*==========================================================================*/
int apci8815_close(int log_slot)
{
	int rslt;

	// åɤƤϽλ
	rslt = apci8815_cancel_int(log_slot);
	if (rslt != APCI8815_OK)
		return rslt;

	// ѥ᡼
	// åȤ
	if (!is_log_slot(log_slot))
		return APCI8815_EINVAL;

	// ǽμ¹Բ
	if (!is_pass(log_slot))
		return APCI8815_EBUSY;

	ctl[log_slot].bexec = APCI8815_FNC_EXEC;		// ¹

	// ǥХΥ
	close(ctl[log_slot].fd);

	ctl[log_slot].fd    = 0;
	ctl[log_slot].bexec = APCI8815_FNC_NOTEXEC;		// ¹Բǽ

	return APCI8815_OK;
}

/*==========================================================================*/
/*  apci8815_set_ad_md - A/D ץ󥰥⡼			    */
/*  Input : log_slot - åȻ					    */
/*	    pmd	     - ץ󥰥⡼			    */
/*  Output: return - 0 ̤ 顼				    */
/*		     0 						    */
/*==========================================================================*/
int apci8815_set_ad_md(int log_slot, papci8815_admd pmd)
{
	// ѥ᡼
	// åȤ
	if (!is_log_slot(log_slot))
		return APCI8815_EINVAL;

	if (pmd == NULL)
		return APCI8815_EINVAL;

	// ǽμ¹Բ
	if (!is_pass(log_slot))
		return APCI8815_EBUSY;

	ctl[log_slot].bexec = APCI8815_FNC_EXEC;		// ¹

	// ioctl ¹
	if (ioctl(ctl[log_slot].fd, APCI8815_SET_AD_MD, pmd) < 0)
	{
		PDEBUG("Error: ioctl, %s\n", strerror(errno));

		ctl[log_slot].bexec = APCI8815_FNC_NOTEXEC;	// ¹Բǽ
		return APCI8815_ESYSCALL;
	}

	ctl[log_slot].bexec = APCI8815_FNC_NOTEXEC;		// ¹Բǽ

	return APCI8815_OK;
}

/*==========================================================================*/
/*  apci8815_start_ad - A/D ץ󥰳				    */
/*  Input : log_slot - åȻ					    */
/*  Output: return - 0 ̤ 顼				    */
/*		     0 						    */
/*==========================================================================*/
int apci8815_start_ad(int log_slot)
{
	// ѥ᡼
	// åȤ
	if (!is_log_slot(log_slot))
		return APCI8815_EINVAL;

	// ǽμ¹Բ
	if (!is_pass(log_slot))
		return APCI8815_EBUSY;

	ctl[log_slot].bexec = APCI8815_FNC_EXEC;		// ¹

	// ioctl ¹
	if (ioctl(ctl[log_slot].fd, APCI8815_START_AD, NULL) < 0)
	{
		PDEBUG("Error: ioctl, %s\n", strerror(errno));

		ctl[log_slot].bexec = APCI8815_FNC_NOTEXEC;	// ¹Բǽ
		return APCI8815_ESYSCALL;
	}

	ctl[log_slot].bexec = APCI8815_FNC_NOTEXEC;		// ¹Բǽ

	return APCI8815_OK;
}

/*==========================================================================*/
/*  apci8815_stop_ad - A/D ץ				    */
/*  Input : log_slot - åȻ					    */
/*  Output: return - 0 ̤ 顼				    */
/*		     0 						    */
/*==========================================================================*/
int apci8815_stop_ad(int log_slot)
{
	// ѥ᡼
	// åȤ
	if (!is_log_slot(log_slot))
		return APCI8815_EINVAL;

	// ǽμ¹Բ
	if (!is_pass(log_slot))
		return APCI8815_EBUSY;

	ctl[log_slot].bexec = APCI8815_FNC_EXEC;		// ¹

	// ioctl ¹
	if (ioctl(ctl[log_slot].fd, APCI8815_STOP_AD, NULL) < 0)
	{
		PDEBUG("Error: ioctl, %s\n", strerror(errno));

		ctl[log_slot].bexec = APCI8815_FNC_NOTEXEC;	// ¹Բǽ
		return APCI8815_ESYSCALL;
	}

	ctl[log_slot].bexec = APCI8815_FNC_NOTEXEC;		// ¹Բǽ

	return APCI8815_OK;
}

/*==========================================================================*/
/*  apci8815_read_ad - A/D ץ󥰥ǡɤ߹			    */
/*  Input : log_slot - åȻ					    */
/*	    buf      - ץ󥰥ǡǼΰ			    */
/*	    count    - ǡ					    */
/*  Output: return - 0 ̤ 顼				    */
/*		     0 ʾ ǡ				    */
/*==========================================================================*/
ssize_t apci8815_read_ad(int log_slot, unsigned short* buf, size_t count)
{
	ssize_t rslt;

	// ѥ᡼
	// åȤ
	if (!is_log_slot(log_slot))
		return APCI8815_EINVAL;

	// ǽμ¹Բ
	if (!is_pass(log_slot))
		return APCI8815_EBUSY;

	ctl[log_slot].bexec = APCI8815_FNC_EXEC;		// ¹

	// ǥХΥ
	rslt = read(ctl[log_slot].fd, buf, count);
	if (rslt < 0)
	{
		PDEBUG("Error: read, %s\n", strerror(errno));

		ctl[log_slot].bexec = APCI8815_FNC_NOTEXEC;	// ¹Բǽ
		return APCI8815_ESYSCALL;
	}

	ctl[log_slot].bexec = APCI8815_FNC_NOTEXEC;		// ¹Բǽ

	return rslt;
}

/*==========================================================================*/
/*  apci8815_get_ad_sts - A/D ץ󥰥ơ		    */
/*  Input : log_slot - åȻ					    */
/*	    psts     - ơǼΰ				    */
/*  Output: return - 0 ̤ 顼				    */
/*		     0 						    */
/*==========================================================================*/
int apci8815_get_ad_sts(int log_slot, unsigned long* psts)
{
	// ѥ᡼
	// åȤ
	if (!is_log_slot(log_slot))
		return APCI8815_EINVAL;

	if (psts == NULL)
		return APCI8815_EINVAL;

	// ǽμ¹Բ
	if (!is_pass(log_slot))
		return APCI8815_EBUSY;

	ctl[log_slot].bexec = APCI8815_FNC_EXEC;		// ¹

	// ioctl ¹
	if (ioctl(ctl[log_slot].fd, APCI8815_GET_AD_STS, psts) < 0)
	{
		PDEBUG("Error: ioctl, %s\n", strerror(errno));

		ctl[log_slot].bexec = APCI8815_FNC_NOTEXEC;	// ¹Բǽ
		return APCI8815_ESYSCALL;
	}

	ctl[log_slot].bexec = APCI8815_FNC_NOTEXEC;		// ¹Բǽ

	return APCI8815_OK;
}

/*==========================================================================*/
/*  apci8815_get_exin - ϥݡȤǡ			    */
/*  Input : log_slot - åȻ					    */
/*	    pexin    - ϥݡȥǡǼΰ			    */
/*  Output: return - 0 ̤ 顼				    */
/*		     0 						    */
/*==========================================================================*/
int apci8815_get_exin(int log_slot, unsigned short* pexin)
{
	// ѥ᡼
	// åȤ
	if (!is_log_slot(log_slot))
		return APCI8815_EINVAL;

	if (pexin == NULL)
		return APCI8815_EINVAL;

	// ǽμ¹Բ
	if (!is_pass(log_slot))
		return APCI8815_EBUSY;

	ctl[log_slot].bexec = APCI8815_FNC_EXEC;		// ¹

	// ioctl ¹
	if (ioctl(ctl[log_slot].fd, APCI8815_GET_EXIN, pexin) < 0)
	{
		PDEBUG("Error: ioctl, %s\n", strerror(errno));

		ctl[log_slot].bexec = APCI8815_FNC_NOTEXEC;	// ¹Բǽ
		return APCI8815_ESYSCALL;
	}

	ctl[log_slot].bexec = APCI8815_FNC_NOTEXEC;		// ¹Բǽ

	return APCI8815_OK;
}

/*==========================================================================*/
/*  apci8815_setup_int - ߥХå롼		    */
/*  Input : log_slot   - åȻ				    */
/*	    pcallback - ߥХå롼Υɥ쥹		    */
/*	    arg	      - Хå롼Ϥ桼(NULL )	    */
/*  Output: return - 0 ̤ 顼				    */
/*		     0 						    */
/*==========================================================================*/
int apci8815_setup_int(int log_slot, void(*pcallback)(int log_slot, void* arg), void* arg)
{
	int rslt;

	// ѥ᡼
	// åȤ
	if (!is_log_slot(log_slot))
		return APCI8815_EINVAL;

	// Хåؿɥ쥹
	if (pcallback == NULL)
		return APCI8815_EINVAL;

	// ˥Хå롼ꤵƤ
	if (ctl[log_slot].hisr != 0)
		return APCI8815_ESISR;			// 

	// ǽμ¹Բ
	if (!is_pass(log_slot))
		return APCI8815_EBUSY;

	ctl[log_slot].bexec = APCI8815_FNC_EXEC;		// ¹

	ctl[log_slot].pcallback = pcallback;			// Хå롼
	ctl[log_slot].arg	   = arg;			// 桼
	ctl[log_slot].exit_isr  = 0;				// åɽλ

	rslt = pthread_create(&ctl[log_slot].hisr, NULL, thread_isr, (void*)log_slot);
	if (rslt)
	{
		PDEBUG("Error: pthread_create, %s\n", strerror(rslt));

		ctl[log_slot].hisr  = 0;			// ǰΤ˽

		ctl[log_slot].bexec = APCI8815_FNC_NOTEXEC;	// ¹Բǽ
		return APCI8815_ESISR;			// 꼺
	}

	ctl[log_slot].bexec = APCI8815_FNC_NOTEXEC;		// ¹Բǽ

	return APCI8815_OK;
}

/*==========================================================================*/
/*  apci8815_cancel_int - ߥХå롼μä	    */
/*  Input : log_slot - åȻ					    */
/*  Output: return - 0 ̤ 顼				    */
/*		     0 						    */
/*==========================================================================*/
int apci8815_cancel_int(int log_slot)
{
	// ѥ᡼
	// åȤ
	if (!is_log_slot(log_slot))
		return APCI8815_EINVAL;

	// Хå롼ꤵƤʤ
	if (ctl[log_slot].hisr == 0)
		return APCI8815_OK;				// ｪλ

	// ǽμ¹Բ
	if (!is_pass(log_slot))
		return APCI8815_EBUSY;

	ctl[log_slot].bexec = APCI8815_FNC_EXEC;		// ¹

	ctl[log_slot].exit_isr = 1;				// åɽλ
	pthread_cancel(ctl[log_slot].hisr);

	ctl[log_slot].hisr = 0;					// åɥϥɥ

	ctl[log_slot].bexec = APCI8815_FNC_NOTEXEC;		// ¹Բǽ

	return APCI8815_OK;
}

/*--------------------------------------------------------------------------*/
/*  dev_open - ǥХΥץ					    */
/*  Input : log_slot - åȻ					    */
/*  Output: return - open 					    */
/*--------------------------------------------------------------------------*/
static int dev_open(int log_slot)
{
	char szbuf[30];

	sprintf(szbuf, "/dev/apci8815%d", log_slot);

	return open(szbuf, O_RDWR);
}

/*--------------------------------------------------------------------------*/
/*  is_log_slot - åȥå					    */
/*  Input : log_slot - åȻ					    */
/*  Output: return - 0 ̵						    */
/*		     1 						    */
/*--------------------------------------------------------------------------*/
static int is_log_slot(int log_slot)
{
	// ϰϤ
	if (!(log_slot >= 0 && log_slot < APCI8815_MAX_DEVICES))
		return 0;

	// ǥХξ
	if (ctl[log_slot].fd == 0)
		return 0;

	// Ѳǽ
	return 1;
}

/*--------------------------------------------------------------------------*/
/*  is_pass - ¹Բǽå						    */
/*  Input : log_slot - åȻ					    */
/*  Output: return - 0 ̵						    */
/*		     1 						    */
/*--------------------------------------------------------------------------*/
static int is_pass(int log_slot)
{
	// ǽμ¹Ծ
	if (ctl[log_slot].bexec != APCI8815_FNC_NOTEXEC)
		return 0;	// ¹Բ

	// ¹Բǽ
	return 1;
}

/*--------------------------------------------------------------------------*/
/*  thread_isr - ߽å					    */
/*  Input : arg - åȻ					    */
/*  Output: return - (void*)0						    */
/*--------------------------------------------------------------------------*/
static void* thread_isr(void* arg)
{
	struct pollfd ufds;

	int log_slot = (int)arg;
	int rslt;

	for (;;)
	{
		// λ郎Ω硢ʲνϤʤ褦ˤ롣
		if (ctl[log_slot].exit_isr)
			break;

		// ߼Ե
		// poll ΣܤΰϡǤ
		// ܤΰϡ-1(ॢ̵)Ǥ
		ufds.fd	    = ctl[log_slot].fd;
		ufds.events = POLLIN;
		rslt = poll(&ufds, 1, -1);
		if (rslt < 0)
		{
			PDEBUG("Error: poll, %s\n", strerror(rslt));
			continue;
		}

		// 󥹥ȡ뤵줿Хå롼θƤӽФ
		ctl[log_slot].pcallback(log_slot, ctl[log_slot].arg);
	}

	return (void*)0;
}
